"""
Some rudimentary tests for queues and broadcasting
"""

# ruff:noqa:D103 pylint: disable=missing-function-docstring
from __future__ import annotations

import anyio
import pytest

from moat.util.broadcast import Broadcaster, LostData

from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from moat.util.broadcast import BroadcastReader


@pytest.mark.anyio()
async def test_example():
    res = []

    async def rdr(bcr: BroadcastReader | Broadcaster, task_status):
        bcr = aiter(bcr)
        task_status.started()
        async for msg in bcr:
            res.append(msg)

    async with anyio.create_task_group() as tg, Broadcaster() as bc:
        await tg.start(rdr, bc)  # "bc" also works
        for x in range(5):
            bc(x)
            await anyio.sleep(0.01)
        bc(42)

    assert res == [0, 1, 2, 3, 4, 42]


@pytest.mark.anyio()
async def test_basic():
    seen = [0, 0, 0]

    async def a(b, n):
        await anyio.sleep(0.1 * (2 * n + 3))
        x = 128
        while True:
            try:
                x = await anext(b)
            except StopAsyncIteration:
                break
            except LostData as exc:
                seen[n] |= exc.n << 4
            else:
                seen[n] |= x

    bq = Broadcaster(1)
    async with anyio.create_task_group() as tg, bq:
        tg.start_soon(a, aiter(bq), 0)
        await anyio.sleep(0.1)
        bq(1)
        await anyio.sleep(0.1)
        tg.start_soon(a, aiter(bq), 1)
        await anyio.sleep(0.1)
        bq(2)
        await anyio.sleep(0.1)
        tg.start_soon(a, aiter(bq), 2)
        await anyio.sleep(0.1)
        bq(4)
        # no delay here
    assert seen == [7, 20, 4]



"""
This module affords a helper class to make "async with OBJECT" map
seamlessly to an async context management method.
"""

from __future__ import annotations


class CtxObj:
    """
    Add an async context manager that calls `_ctx` to run the context.

    Usage::
        class Foo(CtxObj):
            @asynccontextmanager
            async def _ctx(self):
                yield self # or whatever

        async with Foo() as self_or_whatever:
            pass
    """

    async def __aenter__(self):
        self.__ctx = ctx = self._ctx()  # pylint: disable=E1101,W0201
        return await ctx.__aenter__()

    def __aexit__(self, *tb):
        return self.__ctx.__aexit__(*tb)



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

import os
import sys
from collections.abc import Mapping, Sequence

import ruyaml as yaml

from .dict import attrdict
from .msgpack import Proxy
from .path import Path
from .proxy import name2obj

__all__ = ["yload", "yprint", "yformat", "yaml_repr", "yaml_parse", "add_repr"]

SafeRepresenter = yaml.representer.SafeRepresenter
SafeConstructor = yaml.constructor.SafeConstructor
Emitter = yaml.emitter.Emitter


SafeRepresenter.add_representer(attrdict, SafeRepresenter.represent_dict)
SafeConstructor.yaml_base_dict_type = attrdict


def str_presenter(dumper, data):
    """
    Always show multi-line strings with |-style quoting
    """
    if "\n" in data:  # multiline string?
        return dumper.represent_scalar("tag:yaml.org,2002:str", data, style="|")
    else:
        return dumper.represent_scalar("tag:yaml.org,2002:str", data)


def yaml_repr(name: str, use_repr: bool = False):
    """
    A class decorator that allows representing an object in YAML
    """

    def register(cls):
        def str_me(dumper, data):
            return dumper.represent_scalar("!" + name, repr(data) if use_repr else str(data))

        SafeRepresenter.add_representer(cls, str_me)
        return cls

    return register


def yaml_parse(name: str, use_repr: bool = False):
    """
    A decorator that allows parsing a YAML representation,
    i.e. the opposite of `yaml_repr`.
    """

    def register(cls):
        SafeConstructor.add_constructor(f"!{name}", cls)
        return cls

    return register


SafeRepresenter.add_representer(str, str_presenter)


def _path_repr(dumper, data):
    return dumper.represent_scalar("!P", str(data))
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))


def _proxy_repr(dumper, data):
    return dumper.represent_scalar("!R", data.name)
    # return ScalarNode(tag, value, style=style)
    # return yaml.events.ScalarEvent(anchor=None, tag='!P', implicit=(True, True), value=str(data))


def read_env(loader, node):
    value = loader.construct_scalar(node)
    return os.environ[value]


SafeRepresenter.add_representer(Path, _path_repr)
SafeConstructor.add_constructor("!P", Path._make)

SafeConstructor.add_constructor("!env", read_env)

SafeRepresenter.add_representer(Proxy, _proxy_repr)
SafeConstructor.add_constructor("!R", name2obj)


def _bin_from_ascii(loader, node):
    value = loader.construct_scalar(node)
    return value.encode("ascii")


def _bin_from_hex(loader, node):
    value = loader.construct_scalar(node)
    return bytearray.fromhex(value.replace(":", ""))


def _bin_to_ascii(dumper, data):
    try:
        data = data.decode("ascii")
    except UnicodeError:
        if len(data) < 33:
            return dumper.represent_scalar("!hex", data.hex(":"))
        else:
            return dumper.represent_binary(data)
    else:
        return dumper.represent_scalar("!bin", data)


SafeRepresenter.add_representer(bytes, _bin_to_ascii)

SafeConstructor.add_constructor("!bin", _bin_from_ascii)
SafeConstructor.add_constructor("!hex", _bin_from_hex)


_expect_node = Emitter.expect_node


def expect_node(self, *a, **kw):
    """
    YAML stream mangler.

    TODO rationale?
    """
    _expect_node(self, *a, **kw)
    self.root_context = False


Emitter.expect_node = expect_node


def yload(stream, multi=False, attr=False):
    """
    Load one or more YAML objects from a file.
    """
    y = yaml.YAML(typ="safe")
    if attr:

        class AttrConstructor(SafeConstructor):  # pylint: disable=missing-class-docstring
            def __init__(self, *a, **k):
                super().__init__(*a, **k)
                self.yaml_base_dict_type = attrdict if attr is True else attr

        y.Constructor = AttrConstructor
    if multi:
        return y.load_all(stream)
    else:
        return y.load(stream)


def yprint(data, stream=sys.stdout, compact=False):
    """
    Write a YAML record.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    if isinstance(data, (int, float)):
        print(data, file=stream)
    elif isinstance(data, (str, bytes)):
        print(repr(data), file=stream)
    #   elif isinstance(data, bytes):
    #       os.write(sys.stdout.fileno(), data)
    else:
        y = yaml.YAML(typ="safe")
        y.default_flow_style = compact
        y.dump(data, stream=stream)


def yformat(data, compact=None):
    """
    Return ``data`` as a multi-line YAML string.

    :param data: The data to write.
    :param stream: the file to write to, defaults to stdout.
    :param compact: Write single lines if possible, default False.
    """
    from io import StringIO  # pylint: disable=import-outside-toplevel

    s = StringIO()
    yprint(data, compact=compact, stream=s)
    return s.getvalue()


def add_repr(typ, r=None):
    """
    Add a way to add representations for subtypes.

    This is useful for subclassed dict/int/str/… objects.
    """
    # pylint: disable=redefined-builtin
    if r is None:
        r = typ
    if issubclass(r, str):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_str)
    elif issubclass(r, float):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_float)
    elif issubclass(r, bool):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_bool)
    elif issubclass(r, int):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_int)
    elif issubclass(r, Mapping):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_dict)
    elif issubclass(r, Sequence):
        SafeRepresenter.add_representer(typ, SafeRepresenter.represent_list)
    else:
        raise TypeError(f"Don't know what to do with {typ}")



# noqa:D100 pylint: disable=missing-module-docstring,missing-function-docstring
from __future__ import annotations

import pytest


@pytest.fixture()
def anyio_backend():  # noqa:D103
    return "trio"



"""
Testing util.Path
"""

# ruff:noqa:D103 pylint: disable=missing-function-docstring
from __future__ import annotations

import pytest

from moat.util import P, Path, packer, unpacker, yformat, yload

_valid = (
    (("a", "b", "c"), "a.b.c"),
    (("a", 2, "c"), "a:2.c"),
    ((2, "c"), (":i2.c", ":2.c")),
    ((True, "c"), ":t.c"),
    ((1.23, "c"), ":1:.23.c"),
    (("", 1.23, "c"), ":e:1:.23.c"),
    (("a", "", 1.23, "c"), "a:e:1:.23.c"),
    (("a", "", 1.23), "a:e:1:.23"),
    (("a", "", "b"), "a:e.b"),
    (("a", "x y", "b"), ("a.x y.b", "a.x:_y.b")),
    (("a", True), "a:t"),
    (("x", None), "x:n"),
    ((31,), (":x1f", ":31")),
    ((31, "q"), (":x1f.q", ":31.q")),
    (("b", 31, 5), ("b:x1f:5", "b:31:5")),
    (((1, 2), 1.23), (":(1,2):1:.23", ":1,2:1:.23")),
    (((1, 2), "", 1.23), (":(1,2):e:1:.23", ":1,2:e:1:.23")),
    (((1, 2), "c"), ":1,2.c"),
    (((1, "a b", 2), "c"), (":1,'a b',2.c", ":1,'a:_b',2.c")),
    ((), ":"),
    (("a", b"abc"), "a:vabc"),
    (("a", b"ab\x99"), ("a:y616299", "a:sYWKZ")),
    (("a", b"a b"), "a:va:_b"),
    (("a", b"", "c"), "a:v.c"),
)

_invalid = (
    ":w",
    ":t:",
    "a.b:",
    ":2..c",
    "a..b",
    "a.:1",
    "a.:t",
    ":x1g",
    ":x",
    ".a.b",
    "a.b.",
    "a:h123",
    "",
    ":list",
    ":dict",
)


@pytest.mark.parametrize("a,b", _valid)  # noqa:PT006
def test_valid_paths(a, b):
    if isinstance(b, tuple):
        b, xb = b
    else:
        xb = b
    assert str(Path(*a)) == xb
    assert a == tuple(Path.from_str(b))


@pytest.mark.parametrize("a", _invalid)
def test_invalid_paths(a):
    with pytest.raises(SyntaxError):
        Path.from_str(a)


def test_paths():
    p = P("a.b")
    assert str(p) == "a.b"
    q = p | "c"
    assert str(p) == "a.b"
    assert str(q) == "a.b.c"
    r = p + ()  # noqa:RUF005
    assert p is r
    r = p + ("c", "d")  # noqa:RUF005
    assert str(p) == "a.b"
    assert str(r) == "a.b.c.d"
    pp = Path.build(("a", "b"))
    assert str(p) == str(pp)


def test_tagged():
    p = P(":mfoo:")
    assert p.mark == "foo"
    assert len(p) == 0
    p = Path()
    p.mark = "bar"
    assert str(p) == ":mbar:"
    p = P("a:mx.b")
    assert p.mark == "x"  # pylint: disable=no-member
    p = P(":mx.a.b")
    assert p.mark == "x"  # pylint: disable=no-member
    p = P(":mx.a.b:mx")
    assert p.mark == "x"  # pylint: disable=no-member
    p = P("a.b:mx")
    assert p.mark == "x"  # pylint: disable=no-member
    with pytest.raises(SyntaxError):
        P(":mx.a:my.b")
    with pytest.raises(SyntaxError):
        P("a:mx.b:my")


def test_msgpack():
    d = ("a", 1, "b")
    m = packer(d)
    mm = unpacker(m)
    assert type(mm) is tuple  # pylint: disable=unidiomatic-typecheck
    assert mm == d

    d = Path("a", 1, "b")
    m = packer(d)
    mm = unpacker(m)
    assert type(mm) is Path  # pylint: disable=unidiomatic-typecheck
    assert mm == d

    d = {"Hello": d}
    m = packer(d)
    mm = unpacker(m)
    assert type(mm["Hello"]) is Path  # pylint: disable=unidiomatic-typecheck
    assert mm == d


def test_yaml():
    a = Path.from_str("a.b.c")
    b = "!P a.b.c\n...\n"
    assert yformat(a) == b
    assert yload(b) == a



"""
This module contains helper functions for packing+unpacking of single messages,
plus an unpacker factory for streams.

Extension types defined here:
2: contains raw bytes, interpreted as unsigned bignum
3: Path, as a msgpack object stream of its elements
4: contains raw bytes, interpreted as UTF-8, returned as (named) Proxy object
5: object constructor
"""

from __future__ import annotations

from functools import partial

import msgpack

from . import packer, stream_unpacker
from .path import Path
from .proxy import Proxy, _CProxy, obj2name


def _encode(data):
    if isinstance(data, int) and data >= 1 << 64:
        # bignum
        return msgpack.ExtType(2, data.to_bytes((data.bit_length() + 7) // 8, "big"))
    if isinstance(data, Path):
        # Path
        # XXX the mark is dropped until everybody understands type 6
        #   if data.mark:
        #      return msgpack.ExtType(6, packer(data.mark) + b"".join(packer(x) for x in data))
        return msgpack.ExtType(3, b"".join(packer(x) for x in data))
    if isinstance(data, Proxy):
        # Proxy object
        return msgpack.ExtType(5, packer(data.name) + b"".join(packer(x) for x in data.data))
    try:
        name = obj2name(data)
    except KeyError:
        pass
    else:
        return msgpack.ExtType(4, name.encode("utf-8"))

    try:
        name = obj2name(type(data))
    except KeyError:
        pass
    else:
        p = data.__getstate__()
        if not isinstance(p, (list, tuple)):
            p = (p,)
        return msgpack.ExtType(5, packer(name) + b"".join(packer(x) for x in p))

    # XXX we crash instead of sending an unnamed proxy
    # TODO sending a proxied object a second time will build a new one
    return data


def _decode(code, data):
    if code == 2:
        return int.from_bytes(data, "big")
    elif code == 3:
        s = stream_unpacker()
        s.feed(data)
        return Path(*s)
    elif code == 4:
        try:
            n = data.decode("utf-8")
        except UnicodeDecodeError:
            n = str(data)
        try:
            return _CProxy[n]
        except KeyError:
            return Proxy(n)
    elif code == 5:
        s = stream_unpacker()
        s.feed(data)
        s = iter(s)
        pk = next(s)
        try:
            pk = _CProxy[pk]
        except KeyError:
            pk = partial(Proxy, pk)
        pk = object.__new__(pk)
        try:
            pk.__setstate__(*s)
        except AttributeError:
            pk.__dict__.update(next(s))
        return pk
    elif code == 6:
        # A marked path
        s = stream_unpacker()
        s.feed(data)
        s = iter(s)
        mark = next(s)
        p = Path(*s)
        p.mark = mark
        return p
    return msgpack.ExtType(code, data)



"""
Compatibility wrappers that allows MoaT code to run on CPython/anyio as
well as MicroPython/uasyncio.

Well, for the most part.
"""

from __future__ import annotations

import anyio as _anyio
import logging
import os
import time as _time
import traceback as _traceback
from concurrent.futures import CancelledError
from contextlib import suppress
from inspect import currentframe

try:
    import greenback
except ImportError:
    greenback = None

from .queue import Queue as _Queue

logger = logging.getLogger(__name__)

Pin_IN = 0
Pin_OUT = 1

Event = _anyio.Event
Lock = _anyio.Lock
WouldBlock = _anyio.WouldBlock
sleep = _anyio.sleep
EndOfStream = _anyio.EndOfStream
BrokenResourceError = _anyio.BrokenResourceError
ClosedResourceError = _anyio.ClosedResourceError
TimeoutError = TimeoutError  # noqa:PLW0127,A001 pylint:disable=redefined-builtin,self-assigning-variable


class QueueFull(Exception):
    pass


class QueueEmpty(Exception):
    pass


__all__ = [
    "log",
    "Queue",
    "print_exc",
    "ticks_ms",
    "sleep_ms",
    "wait_for",
    "wait_for_ms",
    "every_ms",
    "every",
    "idle",
    "ticks_add",
    "ticks_diff",
    "run",
    "TaskGroup",
    "run_server",
    "shield",
]


def log(s, *x, err=None, nback=1):
    "Basic logger.debug/error call (depends on @err)"
    caller = currentframe()
    for _ in range(nback):
        if caller.f_back is None:
            break
        caller = caller.f_back
    log_ = logging.getLogger(caller.f_globals["__name__"])
    (log_.debug if err is None else log_.error)(s, *x, exc_info=err, stacklevel=1 + nback)
    if err and int(os.getenv("LOG_BRK", "0")):
        breakpoint()  # noqa:T100 pylint:disable=forgotten-debug-statement


def print_exc(exc):
    "print a stack trace to stderr"
    _traceback.print_exception(type(exc), exc, exc.__traceback__)


def ticks_ms():
    "return a monotonic timer, in milliseconds"
    return _time.monotonic_ns() // 1000000


async def sleep_ms(ms):
    "sleep for @ms milliseconds"
    await sleep(ms / 1000)


async def wait_for(timeout, p, *a, **k):
    "timeout if the call to p(*a,**k) takes longer than @timeout seconds"
    with _anyio.fail_after(timeout):
        return await p(*a, **k)


async def wait_for_ms(timeout, p, *a, **k):
    "timeout if the call to p(*a,**k) takes longer than @timeout milliseconds"
    with _anyio.fail_after(timeout / 1000):
        return await p(*a, **k)


async def every_ms(t, p, *a, **k):
    "every t milliseconds, call p(*a,**k)"
    tt = ticks_add(ticks_ms(), t)
    while True:
        try:
            yield await p(*a, **k)
        except StopAsyncIteration:
            return
        tn = ticks_ms()
        td = ticks_diff(tt, tn)
        if td > 0:
            await sleep_ms(td)
            tt += t
        else:
            # owch, delay too long
            tt = ticks_add(tn, t)


def every(t, p, *a, **k):
    "every t seconds, call p(*a,**k)"
    return every_ms(t * 1000, p, *a, **k)


async def idle():
    "sleep forever"
    while True:
        await _anyio.sleep(60 * 60 * 12)  # half a day


def ticks_add(a, b):
    "returns a+b"
    return a + b


def ticks_diff(a, b):
    "returns a-b"
    return a - b


def run(p, *a, **k):
    "wrapper for anyio.run"
    return _anyio.run(p, a, k)


_tg = None


def TaskGroup():
    "A TaskGroup subclass (generator) that supports `spawn` and `cancel`"
    global _tg  # noqa:PLW0603 pylint:disable=global-statement
    if _tg is None:
        _tgt = type(_anyio.create_task_group())

        class TaskGroup_(_tgt):
            """An augmented taskgroup"""

            async def spawn(self, p, *a, _name=None, **k):
                """\
                    Like start(), but returns something you can cancel
                """
                # logger.info("Launch %s %s %s %s",_name, p,a,k)

                async def catch(p, a, k, *, task_status):
                    with _anyio.CancelScope() as s:
                        task_status.started(s)
                        if greenback is not None:
                            await greenback.ensure_portal()
                        with suppress(CancelledError):  # error from concurrent.futures
                            await p(*a, **k)

                return await super().start(catch, p, a, k, name=_name)

            def cancel(self):
                "cancel all tasks in this taskgroup"
                self.cancel_scope.cancel()

        _tg = TaskGroup_
    return _tg()


async def run_server(cb, host, port, backlog=5, taskgroup=None, reuse_port=True, evt=None):
    """Listen to and serve a TCP stream.

    This mirrors [u]asyncio, except that the callback gets the socket once.
    """
    listener = await _anyio.create_tcp_listener(
        local_host=host,
        local_port=port,
        backlog=backlog,
        reuse_port=reuse_port,
    )
    async with listener:
        if evt is not None:
            evt.set()
        await listener.serve(cb, task_group=taskgroup)


def shield():
    """A wrapper shielding the contents from external cancellation.

    Equivalent to ``CancelScope(shield=True)``.
    """
    return _anyio.CancelScope(shield=True)


class Queue(_Queue):
    """
    compatibility mode: raise `EOFError` and `QueueEmpty`/`QueueFull`
    instead of `anyio.EndOfStream` and `anyio.WouldBlock`
    """

    async def get(self):  # noqa:D102
        try:
            return await super().get()
        except _anyio.EndOfStream:
            raise EOFError from None

    def get_nowait(self):  # noqa:D102
        try:
            return super().get_nowait()
        except _anyio.EndOfStream:
            raise EOFError from None
        except _anyio.WouldBlock:
            raise QueueEmpty from None

    def put_nowait(self, x):  # noqa:D102
        try:
            super().put_nowait(x)
        except _anyio.WouldBlock:
            raise QueueFull from None



"""
This module contains proxy helpers.
"""

from __future__ import annotations

from .impl import NotGiven

__all__ = [
    "Proxy",
    "DProxy",
    "NoProxyError",
    "as_proxy",
    "name2obj",
    "obj2name",
    "get_proxy",
    "drop_proxy",
]


class NoProxyError(ValueError):
    "Error for nonexistent proxy values"

    # pylint:disable=unnecessary-pass


class Proxy:
    """
    A proxy object, i.e. a placeholder for things that cannot pass through MsgPack.
    """

    def __init__(self, name):
        self.name = name

    def __repr__(self):
        return f"{self.__class__.__name__}({self.name!r})"

    def ref(self):
        """Dereferences the proxy"""
        return name2obj(self.name)


class DProxy(Proxy):
    """
    A proxy object with data, i.e. an object of a proxied class which
    the receiver doesn't know about
    """

    def __init__(self, name, *a, **k):
        super().__init__(name)
        self.a = a
        self.k = k

    def __getitem__(self, i):
        if i in self.k:
            return self.k[i]
        else:
            return self.a[i]

    def __repr__(self):
        return (
            f"{self.__class__.__name__}({self.name!r},"
            + ",".join(repr(x) for x in (self.a, self.k))
            + ")"
        )


_pkey = 1
_CProxy: dict[str, object] = {}
_RProxy: dict[int, str] = {}


def get_proxy(obj):
    """
    Return a proxy for @obj.

    If no proxy for it exists, a new one is created.
    """
    try:
        return _RProxy[id(obj)]
    except KeyError:
        global _pkey  # noqa:PLW0603 pylint:disable=global-statement
        k = "p_" + str(_pkey)
        _pkey += 1
        _CProxy[k] = obj
        _RProxy[id(obj)] = k
        return k


def drop_proxy(p):
    """
    After sending a proxy we keep it in memory in case the remote returns
    it, or an expression with it.

    If that won't happen, the remote needs to tell us to clean it up.
    """
    if not isinstance(p, str):
        p = _RProxy[id(p)]
    if p == "" or p[0] == "_":
        raise ValueError("Can't delete a system proxy")
    r = _CProxy.pop(p)
    del _RProxy[id(r)]


def name2obj(name, obj=NotGiven, replace=False):
    """
    Translates Proxy name to referred object

    Raises `KeyError` if not found.
    """
    if obj is NotGiven and not replace:
        return _CProxy[name]
    if replace:
        _CProxy[name] = obj
    else:
        oobj = _CProxy.get(name)
        if oobj is not None and oobj is not obj:
            raise KeyError(name)  # exists
    return None


def obj2name(obj, name=NotGiven, replace=False):
    """
    Translates Proxy object to proxied name

    If a name is given, set it.

    Raises `KeyError` if not found and no name given, or if the oid
    already has a different name.
    """
    if name is NotGiven and not replace:
        return _RProxy[id(obj)]
    oid = id(obj)
    if replace:
        _RProxy[oid] = name
    else:
        oname = _RProxy.get(oid)
        if oname is not None and oname != name:
            raise KeyError(name)  # exists
    return None


def as_proxy(name, obj=NotGiven, replace=False):
    """
    Export an object or class as a named proxy.

    @replace can be
    - False (default): error when the name exists
    - True: replace the stored name
    - None: replace the object
    """

    def _proxy(obj):
        if replace is None and name in _CProxy:
            return _CProxy[name]
        _RProxy[id(obj)] = name
        _CProxy[name] = obj
        return obj

    if _CProxy and obj is NotGiven:
        # this allows us to register the NotGiven object
        return _proxy
    else:
        _proxy(obj)
        return obj


as_proxy("_", NotGiven, replace=True)
as_proxy("_p", Proxy)



"""
Some rudimentary tests for queues and broadcasting
"""

# ruff:noqa:D103 pylint: disable=missing-function-docstring
from __future__ import annotations

import anyio
import pytest

from moat.util import AlertCollector


@pytest.mark.anyio()
async def test_collect():
    r = AlertCollector()
    async with r, anyio.create_task_group() as tg:
        busy = False

        async def chk(b):
            nonlocal busy
            await anyio.sleep(0.02)
            assert busy == b
            assert r.is_busy() == b

        async def runner():
            nonlocal busy
            while True:
                await r.wait_busy()
                busy = True
                await r.wait()
                busy = False

        tg.start_soon(runner)
        await anyio.sleep(0.01)
        a = anyio.Event()
        await chk(False)
        r.add(a)
        await chk(True)
        a.set()
        await chk(False)

        a = anyio.Event()
        r.add(a)
        await chk(True)
        b = anyio.Event()
        r.add(b)
        await chk(True)
        a.set()
        await chk(True)
        c = anyio.Event()
        r.add(c)
        await chk(True)
        b.set()
        await chk(True)
        c.set()
        await chk(False)
        tg.cancel_scope.cancel()



"""
More util functions
"""

from __future__ import annotations

from moat.util.misc import pos2val, val2pos


def test_val2pos():
    """
    Translate an input value to the position in an interval
    """
    assert val2pos(0, 1, 2) == 0.5
    assert val2pos(0, 1, 4) == 0.25
    assert val2pos(0, 3, 4) == 0.75
    assert val2pos(4, 3, 0) == 0.25
    assert val2pos(4, 1, 0) == 0.75
    assert val2pos(0, 0, 2) == 0.0
    assert val2pos(0, 2, 2) == 1.0
    assert val2pos(0, -2, 2) == -1.0
    assert val2pos(0, 4, 2) == 2.0
    assert val2pos(0, 4, 2, clamp=True) == 1.0
    assert val2pos(0, -1, 2, clamp=True) == 0.0


def test_pos2val():
    """
    Translate the position in an interval to an output value
    """
    assert pos2val(0, 0.5, 2) == 1
    assert pos2val(0, 0.25, 4) == 1
    assert pos2val(0, 0.75, 4) == 3
    assert pos2val(4, 0.25, 0) == 3
    assert pos2val(4, 0.75, 0) == 1
    assert pos2val(0, 0, 2) == 0
    assert pos2val(0, 1, 2) == 2
    assert pos2val(0, -1, 2) == -2
    assert pos2val(0, 2, 2) == 4
    assert pos2val(0, 1.5, 2, clamp=True) == 2
    assert pos2val(0, -1, 2, clamp=True) == 0.0



"""
Broadcasting support
"""

from __future__ import annotations

try:
    from weakref import WeakSet
except ImportError:
    WeakSet = set

from .compat import EndOfStream, WouldBlock
from .impl import NotGiven
from .queue import Queue

# TODO build something nicer
try:
    import logging
except ImportError:
    logger = None
else:
    logger = logging.getLogger(__name__)

__all__ = [
    "Broadcaster",
    "BroadcastReader",
    "LostData",
]


class LostData(Exception):
    """
    Indicator of data loss.

    Attribute ``n`` contains the number of dropped messages.
    """

    def __init__(self, n):
        self.n = n


class BroadcastReader:
    """
    The read side of a broadcaster.

    Simply iterate over it.

    Warning: The iterator may raise ``LostData`` exceptions.
    These contain the number of messages that have been dropped
    due to the reader being too slow.

    Readers may be called to inject values.
    """

    value = NotGiven
    loss = 0

    def __init__(self, parent, length):
        self.parent = parent
        if length <= 0:
            raise RuntimeError("Length must be at least one")
        self._q = Queue(length)

    def __aiter__(self):
        return self

    async def __anext__(self):
        if self.loss > 0:
            n, self.loss = self.loss, 0
            raise LostData(n)

        try:
            return await self._q.get()
        except (AttributeError, EndOfStream):
            raise StopAsyncIteration from None

    def flush(self):
        """
        Clean the queue.

        Useful for re-sync after you get a `LostData` error.
        """
        try:
            while True:
                self._q.get_nowait()
        except WouldBlock:
            return

    def __call__(self, value):
        """enqueue a value, to this reader only"""
        try:
            self._q.put_nowait(value)
        except WouldBlock:
            x = self._q.get_nowait()
            if logger is not None:
                logger.debug("Dropped: %r", x)
            self._q.put_nowait(value)
            self.loss += 1

    def _close(self):
        self._q.close_writer()

    def close(self):
        "close this reader, detaching it from its parent"
        self._close()
        self.parent._closed_reader(self)  # noqa:SLF001 pylint: disable=protected-access

    async def aclose(self):
        "close this reader, detaching it from its parent"
        self.close()


class Broadcaster:
    """
    A simple broadcaster. Messages will be sent to all readers.

    @length is each reader's default queue length.

    If a queue is full, the oldest message will be discarded. Readers will
    then get a LostData exception that contains the number of dropped
    messages.

    To write, open a context manager (sync or async) and call with a value.

    To read, async-iterate::

        async def rdr(bcr: BroadcastReader|Broadcaster):
            async for msg in bcr:
                print(msg)

        async with anyio.create_task_group() as tg, Broadcaster() as bc:
            tg.start_soon(rdr, aiter(bc))
            for x in range(5):
                bc(x)
                await anyio.sleep(0.01)
            bc(42)

    To safely re-sync, do something like this::

        bcr.flush()
        x = fetch_consistent_state()  # this really should be sync
        bcr(x)  # if this fails, you've got a problem
        while True:
            y = await anext(bcr)
            if x is y:
                break

    """

    _rdr = None
    value = NotGiven

    def __init__(self, length=1):
        self.length = length

    def __repr__(self):
        return f"<{self.__class__.__name__}: {self.length} ({len(self._rdr)})>"

    def open(self):
        """Open the broadcaster.

        Consider using a context instead of this method.
        """
        if self._rdr is not None:
            raise RuntimeError("already entered/opened")
        self._rdr = WeakSet()
        return self

    def __enter__(self):
        return self.open()

    async def __aenter__(self):
        return self.open()

    def __exit__(self, *tb):
        self.close()

    async def __aexit__(self, *tb):
        self.close()

    def _closed_reader(self, reader):
        self._rdr.remove(reader)

    def __aiter__(self):
        """Create a reader with the predefined queue length"""
        r = BroadcastReader(self, self.length)
        self._rdr.add(r)
        return aiter(r)

    def reader(self, length):
        """Create a reader with an explicit queue length"""
        r = BroadcastReader(self, length)
        self._rdr.add(r)
        return aiter(r)

    def __call__(self, value):
        """Enqueue a value to all readers"""
        self.value = value
        for r in self._rdr:
            r(value)

    async def read(self):
        "just gets the value"
        return self.value

    def close(self):
        "Close the broadcaster. No more writing."
        if self._rdr is not None:
            for r in self._rdr:
                r._close()  # noqa:SLF001 pylint: disable=protected-access
            self._rdr = None



"""
Exception handling helpers
"""

from __future__ import annotations

__all__ = ["exc_iter", "ungroup"]


def exc_iter(exc):
    """
    iterate over all non-exceptiongroup parts of an exception(group)
    """
    if isinstance(exc, BaseExceptionGroup):
        for e in exc.exceptions:
            yield from exc_iter(e)
    else:
        yield exc


class ungroup:
    """
    A sync+async context manager that unwraps single-element
    exception groups.
    """
    def __call__(self):
        "Singleton. Returns itself."
        return self

    def one(self, e):
        "convert the exceptiongroup @e to a single exception"
        while isinstance(e, BaseExceptionGroup):
            if len(e.exceptions) != 1:
                break
            e = e.exceptions[0]
        return e

    def __enter__(self):
        return self

    async def __aenter__(self):
        return self

    def __exit__(self, c, e, t):
        if e is None:
            return
        e = self.one(e)
        raise e from None

    async def __aexit__(self, c, e, t):
        return self.__exit__(c, e, t)


ungroup = ungroup()



"""
Some rudimentary tests for packing
"""

# ruff:noqa:D103 pylint: disable=missing-function-docstring
from __future__ import annotations

import pytest

from moat.util import as_proxy, attrdict, packer, unpacker

pytestmark = pytest.mark.skip


class Bar:
    "A proxied object"

    # ruff:noqa:PLW1641

    def __init__(self, x):
        self.x = x

    def __eq__(self, other):
        return self.x == other.x


# needs "replace" because testing re-imports
@as_proxy("fu", replace=True)
class Foo(Bar):
    "A proxied class"

    # pylint: disable=unnecessary-pass


_val = [
    Foo(42),
    attrdict(x=1, y=2),
]


def test_basic():
    for v in _val:
        w = unpacker(packer(v, cbor=True), cbor=True)
        assert v == w, (v, w)


def test_bar():
    b = Bar(95)
    as_proxy("b", b, replace=True)
    c = unpacker(packer(b, cbor=True), cbor=True)
    assert b == c
    with pytest.raises(TypeError):
        packer(Bar(94), cbor=True)



"""
An overly-simple CBOR packer/unpacker.
"""

from __future__ import annotations

import struct
from io import BytesIO

try:
    from micropython import const
except ImportError:

    def const(x):
        "µPy compatibility"
        return x


from moat.util import NotGiven

# Typing
from typing import TYPE_CHECKING  # isort:skip

if TYPE_CHECKING:
    from typing import Any


CBOR_TYPE_MASK = const(0xE0)  # top 3 bits
CBOR_INFO_BITS = const(0x1F)  # low 5 bits


CBOR_UINT = const(0x00)
CBOR_NEGINT = const(0x20)
CBOR_BYTES = const(0x40)
CBOR_TEXT = const(0x60)
CBOR_ARRAY = const(0x80)
CBOR_MAP = const(0xA0)
CBOR_TAG = const(0xC0)
CBOR_7 = const(0xE0)  # float and other types

CBOR_UINT8_FOLLOWS = const(24)  # 0x18
CBOR_UINT16_FOLLOWS = const(25)  # 0x19
CBOR_UINT32_FOLLOWS = const(26)  # 0x1a
CBOR_UINT64_FOLLOWS = const(27)  # 0x1b
CBOR_VAR_FOLLOWS = const(31)  # 0x1f

CBOR_BREAK = const(0xFF)

CBOR_FALSE = const(CBOR_7 | 20)
CBOR_TRUE = const(CBOR_7 | 21)
CBOR_NULL = const(CBOR_7 | 22)
CBOR_UNDEFINED = const(CBOR_7 | 23)  # js 'undefined' value

CBOR_FLOAT16 = const(CBOR_7 | 25)
CBOR_FLOAT32 = const(CBOR_7 | 26)
CBOR_FLOAT64 = const(CBOR_7 | 27)

# CBOR_TAG_DATE_STRING = const(0) # RFC3339
# CBOR_TAG_DATE_ARRAY = const(1) # any number type follows, seconds since 1970-01-01T00:00:00 UTC
CBOR_TAG_BIGNUM = const(2)  # big endian byte string follows
CBOR_TAG_NEGBIGNUM = const(3)  # big endian byte string follows
# CBOR_TAG_DECIMAL = const(4) # [ 10^x exponent, number ]
# CBOR_TAG_BIGFLOAT = const(5) # [ 2^x exponent, number ]
# CBOR_TAG_BASE64URL = const(21)
# CBOR_TAG_BASE64 = const(22)
# CBOR_TAG_BASE16 = const(23)
# CBOR_TAG_CBOR = const(24) # following byte string is embedded CBOR data

# CBOR_TAG_URI = const(32)
# CBOR_TAG_BASE64URL = const()33
# CBOR_TAG_BASE64 = const(34)
# CBOR_TAG_REGEX = const(35)
# CBOR_TAG_MIME = const(36) # following text is MIME message, headers, separators and all
# CBOR_TAG_CBOR_FILEHEADER = const(55799) # can open a file with 0xd9d9f7

_CBOR_TAG_BIGNUM_BYTES = struct.pack("B", CBOR_TAG | CBOR_TAG_BIGNUM)
_CBOR_TAG_NEGBIGNUM_BYTES = struct.pack("B", CBOR_TAG | CBOR_TAG_NEGBIGNUM)


def _dumps_bignum_to_bytearray(val):
    return val.to_bytes((val.bit_length() + 7) // 8, "big")


class ExtraData(ValueError):
    "buffer contains data beyond end of encoded object"

    def __init__(self, unpacked, extra):
        self.unpacked = unpacked
        self.extra = extra

    def __str__(self):
        return "unpack(b) received extra data."


class OutOfData(EOFError):
    "bytes missing"


class Packer:
    "Basic CBOR packer"

    def __init__(self):
        self._buffer = BytesIO()
        # self._unicode_errors = unicode_errors or "strict"

    def _int(self, val):
        "return bytes representing int val in CBOR"
        if val < 0:
            val = -1 - val
            return self._encode_type_num(CBOR_NEGINT, val)
        return self._encode_type_num(CBOR_UINT, val)

    def _bignum_to_bytearray(self, val):
        out = []
        while val > 0:
            out.insert(0, val & 0x0FF)
            val = val >> 8
        return bytes(out)

    def _float(self, val):
        w = self._w
        try:
            ff = struct.pack("!f", val)
        except OverflowError:
            pass
        else:
            if struct.unpack("!f", ff)[0] == val:  # no loss
                try:
                    fe = struct.pack("!e", val)
                except OverflowError:
                    pass
                else:
                    if struct.unpack("!e", fe)[0] == val:  # no loss either
                        return w(struct.pack("!B", CBOR_FLOAT16) + fe)
            return w(struct.pack("!B", CBOR_FLOAT32) + ff)
        return w(struct.pack("!Bd", CBOR_FLOAT64, val))

    def _encode_type_num(self, cbor_type, val) -> None:
        """
        For some CBOR primary type [0..7] and an auxiliary unsigned
        number, return CBOR encoded bytes
        """
        w = self._w

        if val is None:
            return w(struct.pack("B", cbor_type | CBOR_VAR_FOLLOWS))

        assert val >= 0
        if val <= 23:
            return w(struct.pack("B", cbor_type | val))
        if val <= 0x0FF:
            return w(struct.pack("BB", cbor_type | CBOR_UINT8_FOLLOWS, val))
        if val <= 0x0FFFF:
            return w(struct.pack("!BH", cbor_type | CBOR_UINT16_FOLLOWS, val))
        if val <= 0x0FFFFFFFF:
            return w(struct.pack("!BI", cbor_type | CBOR_UINT32_FOLLOWS, val))
        if val <= (0x07FFFFFFFFFFFFFFF if cbor_type == CBOR_NEGINT else 0x0FFFFFFFFFFFFFFFF):
            return w(struct.pack("!BQ", cbor_type | CBOR_UINT64_FOLLOWS, val))

        # too large: write a bignum
        if cbor_type == CBOR_UINT:
            w(_CBOR_TAG_BIGNUM_BYTES)
        elif cbor_type == CBOR_NEGINT:
            w(_CBOR_TAG_NEGBIGNUM_BYTES)
        else:
            raise ValueError("value too big: " + repr(val))
        outb = _dumps_bignum_to_bytearray(val)
        self._encode_type_num(CBOR_BYTES, len(outb))
        w(self._bignum_to_bytearray(val))
        return None

    def _string(self, val):
        if isinstance(val, str):
            val = val.encode("utf8")
            self._encode_type_num(CBOR_TEXT, len(val))
        else:
            self._encode_type_num(CBOR_BYTES, len(val))
        self._w(val)

    def _array(self, arr):
        try:
            self._encode_type_num(CBOR_ARRAY, len(arr))
        except TypeError:
            self._encode_type_num(CBOR_ARRAY, None)
            for x in arr:
                self._any(x)
            self._w(struct.pack("B", CBOR_BREAK))
        else:
            for x in arr:
                self._any(x)

    def _dict(self, d):
        self._encode_type_num(CBOR_MAP, len(d))
        for k, v in d.items():
            self._any(k)
            self._any(v)

    def _bool(self, b):
        if b:
            self._w(struct.pack("B", CBOR_TRUE))
        self._w(struct.pack("B", CBOR_FALSE))

    def _tag(self, t):
        self._encode_type_num(CBOR_TAG, t.tag)
        self._any(t.value)

    def _w(self, d):
        self._buffer.write(d)

    def _any(self, ob):
        w = self._w
        if ob is None:
            w(struct.pack("B", CBOR_NULL))
        elif ob is NotGiven:
            w(struct.pack("B", CBOR_UNDEFINED))
        elif isinstance(ob, bool):
            self._bool(ob)
        elif isinstance(ob, (str, bytes)):
            self._string(ob)
        elif isinstance(ob, (list, tuple)):
            self._array(ob)
        # TODO: accept other enumerables and emit a variable length array
        elif isinstance(ob, dict):
            self._dict(ob)
        elif isinstance(ob, float):
            self._float(ob)
        elif isinstance(ob, int):
            self._int(ob)
        elif isinstance(ob, Tag):
            self._tag(ob)
        else:
            raise TypeError("Cannot serialize type " + repr(type(ob)))

    def packb(self, obj):
        "pack @obj, return the resulting bytes"
        if self._buffer.getvalue():
            raise RuntimeError("Writer busy")
        try:
            self._any(obj)
            return self._buffer.getvalue()
        finally:
            self._buffer = BytesIO()  # always reset


class Tag:
    "a random CBOR tag"

    def __init__(self, tag=None, value=None):
        self.tag = tag
        self.value = value

    def __repr__(self):
        return f"Tag({self.tag !r}, {self.value !r})"

    def __eq__(self, other):
        if not isinstance(other, Tag):
            return False
        return self.tag == other.tag and self.value == other.value

    def __hash__(self):
        return hash(self.tag) ^ hash(self.value)


class Unpacker:
    "Basic CBOR unpacker"

    _stream = None
    _buff_i = 0
    _rsz = 64

    _MAX_DEPTH = 100

    def __init__(self, stream=None, tag_hook=Tag):
        self._stream = stream
        self._buffer = bytearray()
        self.tag_hook = tag_hook

    async def _tag_aux(self, tb):
        tag = tb & CBOR_TYPE_MASK
        tag_aux = tb & CBOR_INFO_BITS
        if tag_aux <= 23:
            aux = tag_aux
        elif tag_aux == CBOR_UINT8_FOLLOWS:
            data = await self._read(1)
            aux = struct.unpack_from("!B", data, 0)[0]
        elif tag_aux == CBOR_UINT16_FOLLOWS:
            data = await self._read(2)
            aux = struct.unpack_from("!H", data, 0)[0]
        elif tag_aux == CBOR_UINT32_FOLLOWS:
            data = await self._read(4)
            aux = struct.unpack_from("!I", data, 0)[0]
        elif tag_aux == CBOR_UINT64_FOLLOWS:
            data = await self._read(8)
            aux = struct.unpack_from("!Q", data, 0)[0]
        else:
            assert tag_aux == CBOR_VAR_FOLLOWS, f"bogus tag {tb:02x}"
            aux = None

        return tag, aux

    async def _read(self, n):
        # (int) -> bytearray
        await self._reserve(n)
        i = self._buff_i
        ret = self._buffer[i : i + n]
        self._buff_i = i + len(ret)
        return ret

    async def _read_byte(self):
        return (await self._read(1))[0]

    async def _reserve(self, n):
        nb = self._buff_i + n - len(self._buffer)

        # Fast path: buffer has n bytes already
        if nb <= 0:
            return

        if not self._stream:
            raise OutOfData

        # Read from file
        while nb > 0:
            to_read_bytes = max(self._rsz, nb)
            read_data = await self._stream.recv(to_read_bytes)
            if not read_data:
                break
            self._buffer += read_data
            nb -= len(read_data)

        if len(self._buffer) < n + self._buff_i:
            self._buff_i = 0  # rollback
            raise OutOfData

    def __aiter__(self):
        return self

    def __anext__(self):
        return self.unpack()

    def __iter__(self):
        return self

    def __next__(self):
        g = self.unpack()
        try:
            g.send(None)
        except StopIteration as exc:
            return exc.value
        except OutOfData:
            raise StopIteration from None
        #       except BaseException as err:
        #           raise RuntimeError(err)
        else:
            raise RuntimeError("Needs async")

    async def _var_array(self):
        ob = []
        tb = await self._read_byte()
        while tb != CBOR_BREAK:
            subob = await self._tb(tb)
            ob.append(subob)
            tb = await self._read_byte()
        return ob

    async def _var_map(self):
        ob = {}
        tb = self._read_byte()
        while tb != CBOR_BREAK:
            subk = await self._tb(tb)
            subv = await self._any()
            ob[subk] = subv
            tb = self._read_byte()
        return ob

    async def _array(self, aux):
        ob = []
        for _ in range(aux):
            subob = await self._any()
            ob.append(subob)
        return ob

    async def _map(self, aux):
        ob = {}
        for _ in range(aux):
            subk = await self._any()
            subv = await self._any()
            ob[subk] = subv
        return ob

    async def unpack(self) -> Any:
        "unpack my buffer into an object"
        res = await self._any()
        # Buffer management: chop off the part we've read
        self._buffer = self._buffer[self._buff_i :]
        self._buff_i = 0
        return res

    def _got_extradata(self):
        return self._buff_i < len(self._buffer)

    def _get_extradata(self):
        return self._buffer[self._buff_i :]

    def unpackb(self, packed) -> Any:
        "unpack"
        self.feed(packed)
        try:
            upack = self.unpack()
            upack.send(None)
        except StopIteration as s:
            if self._got_extradata():
                raise ExtraData(s.value, bytes(self._get_extradata())) from None
            return s.value
        except OutOfData:
            raise ValueError("incomplete") from None
        raise RuntimeError("No way")

    def feed(self, data: bytes):
        "set the decode buffer"
        assert self._stream is None
        self._buffer = memoryview(data)
        self._buff_i = 0

    async def _any(self):
        #       if depth > _MAX_DEPTH:
        #           raise RuntimeError("load recursion limit")

        tb = await self._read_byte()
        return await self._tb(tb)

    async def _tb(self, tb):
        # Some special cases of CBOR_7 best handled by special struct.unpack logic here
        if tb == CBOR_FLOAT16:
            data = await self._read(2)
            pf = struct.unpack_from("!e", data, 0)
            return pf[0]
        elif tb == CBOR_FLOAT32:
            data = await self._read(4)
            pf = struct.unpack_from("!f", data, 0)
            return pf[0]
        elif tb == CBOR_FLOAT64:
            data = await self._read(8)
            pf = struct.unpack_from("!d", data, 0)
            return pf[0]

        tag, aux = await self._tag_aux(tb)

        if tag == CBOR_UINT:
            return aux
        elif tag == CBOR_NEGINT:
            return -1 - aux
        elif tag == CBOR_BYTES:
            return await self._bytes(aux)
        elif tag == CBOR_TEXT:
            return str(await self._bytes(aux, btag=CBOR_TEXT), "utf8")
        elif tag == CBOR_ARRAY:
            if aux is None:
                return await self._var_array()
            return await self._array(aux)
        elif tag == CBOR_MAP:
            if aux is None:
                return await self._var_map()
            return await self._map(aux)
        elif tag == CBOR_TAG:
            ob = await self._any()
            # attempt to interpet the tag and the value into a Python object.
            return self.tag_hook(aux, ob)
        else:
            assert tag == CBOR_7
            if tb == CBOR_TRUE:
                return True
            if tb == CBOR_FALSE:
                return False
            if tb == CBOR_NULL:
                return None
            if tb == CBOR_UNDEFINED:
                return NotGiven
            raise ValueError(f"unknown cbor tag 7 byte: {tb:02x}")

    async def _bytes(self, aux, btag=CBOR_BYTES):
        # TODO: limit to some maximum number of chunks and some maximum total bytes
        if aux is not None:
            # simple case
            ob = await self._read(aux)
            return ob
        # read chunks of bytes
        chunklist = []
        while True:
            tb = await self._read_byte()
            if tb == CBOR_BREAK:
                break
            tag, aux = await self._tag_aux(tb)
            if tag != btag:
                raise ValueError("var length with unexpected component")
            ob = await self._read(aux)
            chunklist.append(ob)
        return b"".join(chunklist)


def tagify(aux, ob):
    "create data types from tagged CBOR input"
    # TODO: make this extensible?
    # cbor.register_tag_handler(tagnumber, tag_handler)
    # where tag_handler takes (tagnumber, tagged_object)
    #   if aux == CBOR_TAG_DATE_STRING:
    #       # TODO: parse RFC3339 date string
    #       pass
    #   if aux == CBOR_TAG_DATE_ARRAY:
    #       return datetime.datetime.utcfromtimestamp(ob)
    if aux == CBOR_TAG_BIGNUM:
        return int.from_bytes(ob, "big")
    if aux == CBOR_TAG_NEGBIGNUM:
        return -1 - int.from_bytes(ob, "big")
    # if aux == CBOR_TAG_REGEX:
    # Is this actually a good idea? Should we just return
    # the tag and the raw value to the user somehow?
    #       return re.compile(ob)
    return Tag(aux, ob)


def packb(o, **kwargs):
    """
    Pack object `o` and return packed bytes

    See :class:`Packer` for options.
    """
    return Packer(**kwargs).packb(o)


def unpackb(packed, **kwargs):
    """
    Unpack an object from `packed`.

    See :class:`Unpacker` for options.
    """
    unpacker = Unpacker(None, **kwargs)
    return unpacker.unpackb(packed)



"""
This module contains a heap of somewhat-random helper functions
and classes which are used throughout MoaT (and beyond)
but don't get their own package because they're too small,
or too interrelated … or the author was too lazy.
"""
# TODO split this up

# pylint: disable=cyclic-import,wrong-import-position
from __future__ import annotations

import logging as _logging

import msgpack as _mp

_log = _logging.getLogger(__name__)


def packer(*a, cbor=False, **k):
    """single message packer"""
    if cbor:
        return _cbor.packb(*a, **k)
    # ruff:noqa:SLF001 pylint:disable=protected-access
    return _mp.packb(*a, strict_types=False, use_bin_type=True, default=_msgpack._encode, **k)


def unpacker(*a, cbor=False, **k):
    """single message unpacker"""
    if cbor:
        return _cbor.unpackb(*a, **k)
    return _mp.unpackb(
        *a,
        object_pairs_hook=attrdict,
        strict_map_key=False,
        raw=False,
        use_list=False,
        ext_hook=_msgpack._decode,  # pylint:disable=protected-access
        **k,
    )


def stream_unpacker(*a, cbor=False, **k):
    """stream unpacker factory"""
    if cbor:
        return _cbor.Unpacker(*a, **k)
    return _mp.Unpacker(
        *a,
        object_pairs_hook=attrdict,
        strict_map_key=False,
        raw=False,
        use_list=False,
        ext_hook=_msgpack._decode,  # pylint:disable=protected-access
        **k,
    )


from .dict import attrdict  # noqa: E402

from .alert import *  # noqa: F403, E402  # isort:skip
from .impl import *  # noqa: F403, E402  # isort:skip
from .dict import *  # noqa: F403, E402  # isort:skip
from .merge import *  # noqa: F403, E402  # isort:skip
from .proxy import *  # noqa: F403, E402  # isort:skip
from .misc import *  # noqa: F403, E402  # isort:skip

try:
    from .event import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .ctx import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .queue import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .module import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .msg import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .path import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .server import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .spawn import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .systemd import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .yaml import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .exc import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

try:
    from .main import *  # noqa: F403
except ImportError as exc:
    _log.warning("Missing: %s", exc)

from . import cbor as _cbor  # noqa:E402
from . import msgpack as _msgpack  # noqa:E402 pylint:disable=reimported  # nonsense



"""
Basic tool support
"""

from __future__ import annotations

import anyio
import logging  # pylint: disable=wrong-import-position
import sys
from datetime import datetime
from time import time

import asyncclick as click

from . import packer, stream_unpacker
from .main import load_subgroup
from .path import P, Path, path_eval
from .times import humandelta, time_until
from .yaml import yload, yprint

log = logging.getLogger()


@load_subgroup(prefix="moat.util")
async def cli():
    """Various utilities"""
    # pylint:disable=unnecessary-pass


@cli.command(name="to")
@click.option("--sleep", "-s", is_flag=True, help="Sleep until that date/time")
@click.option("--human", "-h", is_flag=True, help="Print in numan-readable terms")
@click.option("--now", "-n", is_flag=True, help="Don't advance on match")
@click.option("--inv", "-i", is_flag=True, help="Time until no match")
@click.option("--back", "-b", is_flag=True, help="Time since the last (non)-match")
@click.argument("args", nargs=-1)
async def to_(args, sleep, human, now, inv, back):
    """\
Calculate the time until the start of the next given partial time
specification.

For instance, "moat util to 9 h": shows in how many seconds it's 9 o'clock
(possibly on the next day). Arbitrarily many units can be combined.

Negative numbers count from the end, i.e. "-2 hr" == 10 pm. Don't
forget to use the "--" option-to-argument separator if the time
specification starts with a negative number.

Days are numbered 1…7, Monday…Sunday. "3 dy" is synonymous to "wed",
while "3 wed" means "the third wednesday in a month".

"--human" prints a human-understandable version of the given
time. "--sleep" then delays until the specified moment arrives. If none
of these options is given, the number of seconds is printed.

By default, if the given time spec matches the current time, the
duration to the *next* moment the spec matches is calculated. Use
"--now" to print 0 / "now" / not sleep instead.

"--inv" inverts the time specification, i.e. "9 h" prints the time
until the next moment it is not / again no longer 9:** o'clock,
depending on whether "--now" is used / not used.

"--back" calculates the time *since the end* of the last match /
non-match instead. (If you want the start, use "--inv" and add a
second.)

\b
Known units:
s, sec  Second (0…59)
m, min  Minute (0…59)
h, hr   Hour (0…23)
d, dy   Day-of-week (1…7)
mon…sun Day-in-month (1…5)
w, wk   Week-of-year (0…53)
m, mo   Month (1…12)
y, yr   Year (2023–)
"""
    if not args:
        raise click.UsageError("Up to when please?")
    if back and sleep:
        raise click.UsageError("We don't have time machines.")

    t = datetime.now().astimezone()
    if not now:
        t = time_until(args, t, invert=not inv, back=back)
    t = time_until(args, t, invert=inv, back=back)

    t = t.timestamp()
    tt = int(t - time() + 0.9)
    if back:
        tt = -tt
    if human:
        print(humandelta(tt))
    if sleep:
        await anyio.sleep(tt)
    elif not human:
        print(tt)


@cli.command
@click.option("-d", "--decode", is_flag=True, help="decode (default: encode)")
@click.argument("path", type=click.Path(file_okay=True, dir_okay=False))
def msgpack(decode, path):
    """encode/decode msgpack from/to YAML

    moat util msgpack data.yaml > data.mp    # encode
    moat util msgpack -d data.mp > data.yaml # decode
    """
    if decode:
        with sys.stdin.buffer if path == "-" else open(path, "rb") as f:
            n = 0
            for obj in stream_unpacker(f):
                if n:
                    print("---")
                n += 1
                yprint(obj)
    else:
        with sys.stdin if path == "-" else open(path) as f:
            for obj in yload(f, multi=True):
                sys.stdout.buffer.write(packer(obj))


@cli.command("path", help=Path.__doc__, no_args_is_help=True)
@click.option(
    "-e",
    "--encode",
    is_flag=True,
    help="evaluate a Python expr and encode to a pathstr",
)
@click.option("-d", "--decode", is_flag=True, help="decode a path to a list")
@click.argument("path", nargs=-1)
async def path_(encode, decode, path):
    """Explain/test MoaT paths"""
    if not encode and not decode:
        raise click.UsageError("Need -e or -d option.")
    elif not decode:
        try:
            path = path_eval(" ".join(path))
        except Exception as exc:  # pylint:disable=broad-exception-caught
            print(repr(exc), file=sys.stderr)
        if not isinstance(path, (list, tuple)):
            path = path[0]
        print(Path(*path))
    elif encode:
        raise click.UsageError("encode and decode at the same time??")
    else:
        for p in path:
            print(repr(list(P(p))))


@cli.command("cfg", help="Retrieve+show a config value", no_args_is_help=True)
@click.pass_obj
@click.option("-y", "--yaml", is_flag=True, help="print as YAML")
@click.option("-e", "--empty", is_flag=True, help="empty string if the key doesn't exist")
@click.argument("path", nargs=-1, type=P)
async def cfg_(obj, path, yaml, empty):
    """Emit the current configuration as a YAML file.

    You can limit the output by path elements.
    E.g., "cfg kv.connect.host" will print "localhost".

    Single values are printed with a trailing line feed.

    Dump the whole config with "moat util cfg :".
    """
    from .exc import ungroup

    delim = False
    for p in path:
        if delim and yaml:
            print("---", file=obj.stdout)
        with ungroup():
            try:
                v = obj.cfg._get(p)  # noqa:SLF001
            except KeyError:
                if not empty:
                    print("Unknown:", p, file=sys.stderr)
                    sys.exit(1)
            else:
                if yaml:
                    yprint(v, obj.stdout)
                else:
                    print(v, file=obj.stdout)
        delim = True



"""
This module contains various helper functions and classes.
"""

from __future__ import annotations

__all__ = ["val2pos", "pos2val"]


def val2pos(a: float, b: float, c: float, /, clamp: bool = False):
    """
    Return the position of @b within [@a…@c].

        b==a => 0
        b==c => 1
        b==(a+c)/2 => 0.5

    If @clamp is set, the return value is limited to [0…1] even if @b is
    outside a…c.

    @a and @c may not be equal.
    """

    res = (a - b) / (a - c)
    if clamp:
        if res < 0:
            res = 0
        elif res > 1:
            res = 1
    return res


def pos2val(a: float, b: float, c: float, /, clamp: bool = False):
    """
    Return the value of @b within [@a…@c].

        a,0,c => a
        a,1,c => c
        a,0.5,c => (a+c)/2
        a,-1,c => 2*a-c
        a,2,c => 2*c-a

    If @clamp is set, the input is limited to [0…1].

    @a and @c may not be equal.
    """

    if clamp:
        if b < 0:
            b = 0
        elif b > 1:
            b = 1

    return a + b * (c - a)
